from pathlib import Path
from typing import Annotated, Any, Self

from pydantic import BaseModel, BeforeValidator, SerializeAsAny, field_validator, model_validator

from contextcheck.assertions.assertions import AssertionBase
from contextcheck.assertions.factory import factory as assertions_factory
from contextcheck.endpoints.endpoint_config import EndpointConfig
from contextcheck.loaders.yaml import load_yaml_file
from contextcheck.models.request import RequestBase
from contextcheck.models.response import ResponseBase


class TestConfig(BaseModel):
    endpoint_under_test: EndpointConfig = EndpointConfig()
    default_request: RequestBase | None = None
    eval_endpoint: EndpointConfig | None = None


class TestStep(BaseModel):
    name: str
    request: RequestBase
    response: ResponseBase | None = None
    default_request: RequestBase = RequestBase()
    asserts: list[SerializeAsAny[AssertionBase]] = []
    result: bool | None = None

    @model_validator(mode="before")
    @classmethod
    def from_obj(cls, obj: dict | str) -> dict:
        # Default test step is request with `message` field
        return obj if isinstance(obj, dict) else {"name": obj, "request": RequestBase(message=obj)}

    @model_validator(mode="after")
    def use_default_request(self) -> Self:
        if self.default_request:
            self.request = self.default_request.model_copy(update=self.request.model_dump())
        return self

    @field_validator("asserts")
    @classmethod
    def prepare_asserts(cls, asserts: list[AssertionBase]) -> list[AssertionBase]:
        prepared_asserts = [assertions_factory(assert_.model_dump()) for assert_ in asserts]
        return prepared_asserts


class TestScenario(BaseModel):
    __test__ = False
    steps: list[TestStep] = []
    config: Annotated[TestConfig, BeforeValidator(lambda x: {} if x is None else x)]
    result: bool | None = None
    filename: str | None = None

    @model_validator(mode="before")
    @classmethod
    def provide_default_request_for_steps(cls, data: Any) -> Any:
        default_request = None
        steps = None
        if isinstance(data, BaseModel):
            data = data.model_dump()

        if isinstance(data, dict):
            if config := data.get("config", {}):
                if isinstance(config, dict) and config.get("default_request", {}):
                    default_request = config.get("default_request")

            steps = data.get("steps", [])

        if default_request is not None and isinstance(steps, list) and steps:
            new_steps = []
            for step in steps:
                new_step = step
                if isinstance(step, str):
                    new_step = {"name": step, "request": step, "default_request": default_request}
                elif isinstance(step, dict):
                    step["default_request"] = step.get("default_request", {}) or default_request
                new_steps.append(new_step)

            data["steps"] = new_steps
        # We can return the same data, as everything was in place
        return data

    @classmethod
    def from_yaml(cls, file_path: Path | str) -> Self:
        file_path = Path(file_path)
        cls_dict = load_yaml_file(file_path=file_path)
        cls_dict["filename"] = file_path.name
        return cls.model_validate(cls_dict)

    def show_test_step_results(self):
        # NOTE: For better visual aspects we could check rich table
        print("-" * 12)
        for step in self.steps:
            print(f"Name: {step.name}; Result: {step.result}\n")
            for assertion in step.asserts:
                assertion_dumped = assertion.model_dump()
                assertion_ = assertion.eval if "eval" in assertion_dumped else assertion.assertion
                print(f'Assertion: "{assertion_}", Result: {assertion.result}')
            print("-" * 12)



from contextcheck.connectors.connector_echo import ConnectorEcho
from contextcheck.endpoints.endpoint import EndpointBase


class EndpointDummyEcho(EndpointBase):
    connector: ConnectorEcho = ConnectorEcho()



import argparse

from contextcheck.executors.tests_router import TestsRouter


class CustomArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        self.print_help()
        self.exit(2, f"\n{self.prog}: error: {message}\n")

    def validate_args(self, args):
        if args.output_type == "file":
            if not args.filename and not args.folder:
                self.error("--filename or --folder is required")
            if (args.filename or args.folder) and not args.output_folder:
                self.error(
                    '--output-folder is required when --output-type is "file" and filename or folder is set'
                )


def main():
    parser = CustomArgumentParser(
        description=(
            "ContextCheck Tests Runner. Run tests from a file or folder and output the results in the console or save to JSON files."
        )
    )

    parser.add_argument(
        "--output-type",
        required=True,
        choices=["console", "file"],
        help='Specify the output type: "console" or "file"',
    )

    parser.add_argument(
        "--filename",
        nargs="+",
        help="Filename (or list of space separated filenames) to use for tests",
    )

    parser.add_argument("--folder", help="Folder name to load all YAML files from for tests")

    parser.add_argument(
        "--output-folder",
        help='[Required if --output-type is "file"] Specify the output folder to save the test results in JSON format',
    )

    parser.add_argument(
        "--exit-on-failure", action="store_true", help="Exit with status code 1 if any test fails"
    )

    parser.add_argument(
        "--aggregate-results", action="store_true", help="Aggregate assertion results"
    )

    parser.add_argument(
        "--show-time-statistics", action="store_true", help="Show aggregated time statistics"
    )

    args = parser.parse_args()
    parser.validate_args(args)

    TestsRouter(**vars(args)).run_tests()


if __name__ == "__main__":
    main()



from contextcheck.assertions.assertions import (
    AssertionBase,
    AssertionDeterministic,
    AssertionEval,
    AssertionLLM,
)
from contextcheck.assertions.settings import AssertionKind

assertions_map = {
    AssertionKind.EVAL: AssertionEval,
    AssertionKind.LLM_METRIC: AssertionLLM,
    AssertionKind.DETERMINISTIC: AssertionDeterministic,
}


def factory(assert_definition: dict) -> AssertionBase:
    # Take first key from assertions_map present in assert_deffinition
    try:
        kind = next(
            assert_key for assert_key in assert_definition.keys() if assert_key in assertions_map
        )
    except StopIteration:
        raise ValueError(f"No assertion for definition {assert_definition}")

    assertion_class = assertions_map[kind]
    return assertion_class.model_validate(assert_definition)



from pathlib import Path
from typing import Any

from pydantic import BaseModel
from rich import print
from rich.console import Group, RenderableType
from rich.panel import Panel
from rich.pretty import Pretty
from rich.table import Table
from rich.text import Text

from contextcheck.assertions.assertions import AssertionBase
from contextcheck.executors.executor import Executor
from contextcheck.interfaces.interface import InterfaceBase
from contextcheck.models.models import TestStep
from contextcheck.models.request import RequestBase
from contextcheck.models.response import ResponseBase


def _create_panel(text: str, obj: BaseModel, width: int = 80) -> RenderableType:
    return Panel(
        Group(
            Text(text, style="bold red"),
            Pretty(obj),
        ),
        width=width,
    )


class InterfaceTUI(InterfaceBase):
    test_scenario_filename: str | None = None

    def get_scenario_path(self) -> Path:
        return Path(self.test_scenario_filename or "")

    def __call__(self, obj: BaseModel) -> Any:
        if isinstance(obj, TestStep):
            print(obj)
            return
        elif isinstance(obj, ResponseBase):
            text = "💬 Response:"
        elif isinstance(obj, RequestBase):
            text = "🎈 Request:"
        elif isinstance(obj, AssertionBase):
            text = "🧐 Assertion:"
        else:
            text = "Unknown"
        print(_create_panel(text, obj))

    def summary(self, executor: Executor, **kwargs: Any) -> None:
        table = Table(show_lines=True)
        table.add_column("Request")
        table.add_column("Response")
        table.add_column("Asserts")
        table.add_column("Valid")

        for step in executor.test_scenario.steps:
            step_result = ""
            if step.result is None:
                step_result = "[yellow]SKIPPED"
            elif step.result:
                step_result = "[green]OK"
            else:
                step_result = "[red]FAIL"

            table.add_row(
                Pretty(step.request),
                Pretty(step.response),
                Pretty(step.asserts),
                step_result,
            )

        print(table)

        if kwargs.get("aggregate_results", False):
            self.report_results(executor=executor)

        if kwargs.get("show_time_statistics", False):
            self.report_time(executor=executor)

    def report_results(self, executor: Executor, **kwargs) -> None:
        scenario_results = self._create_a_summary_report(executor=executor)

        # NOTE: Now we have only binary statistics, but if we were to add "continous assertions"
        # then we'd need to either update this table, or create a separate table for continous results
        # Although we could also add some new overlapping metrics like max, min, median etc.
        table = Table(show_lines=True)
        table.add_column("Metric type")  # LLM Metric, eval etc.
        table.add_column("Metric name")  # Name of the metric e.g. LLM-Metric / qa-reference
        table.add_column("Mean score")
        table.add_column("Test count")

        for key, value in scenario_results.items():
            for key2, value2 in value.items():
                table.add_row(
                    Pretty(key),
                    Pretty(key2),
                    Pretty(value2["mean"]),
                    Pretty(value2["count"]),
                )

        print(table)

    def report_time(self, executor: Executor, **kwargs) -> None:
        time_statistics = self._create_time_statistics(executor=executor)

        # NOTE: Now we have only binary statistics, but if we were to add "continous assertions"
        # then we'd need to either update this table, or create a separate table for continous results
        # Although we could also add some new overlapping metrics like max, min, median etc.
        table = Table(show_lines=True)
        table.add_column("Mean")  # LLM Metric, eval etc.
        table.add_column("Median")  # Name of the metric e.g. LLM-Metric / qa-reference
        table.add_column("Minimum")
        table.add_column("Maximum")
        table.add_column("Std")

        table.add_row(
            Pretty(time_statistics["mean"]),
            Pretty(time_statistics["median"]),
            Pretty(time_statistics["minimum"]),
            Pretty(time_statistics["maximum"]),
            Pretty(time_statistics["std"]),
        )

        print(table)



from contextcheck.connectors.connector_http import ConnectorHTTP
from contextcheck.endpoints.endpoint import EndpointBase


class EndpointTGChatBot(EndpointBase):

    def model_post_init(self, __context) -> None:
        self.connector = ConnectorHTTP(
            url=self.config.url,
            additional_headers=self.config.additional_headers,
        )



from pathlib import Path

import yaml
from jinja2 import Template


def load_yaml_file(file_path: Path, parse_template: bool = True) -> dict:
    """Load yaml file and optionally parse it before as jinja2 template."""
    if not file_path.exists():
        raise FileNotFoundError(f"File '{file_path}' does not exist.")

    with open(file_path, "r") as file:
        yaml_content = file.read()

    if parse_template:
        # Get variables from original yaml:
        try:
            yaml_dict_without_template = yaml.safe_load(yaml_content)
        except yaml.scanner.ScannerError as e: # type: ignore
            message = ("Yaml parsing error. It was probably caused by your 'regex' assertion " 
                f"defined with double-quotes.\nUse single quotes or no quotes to fix it.\nError details:\n\n{e}")
            raise ValueError(message)
            
        
        variables = yaml_dict_without_template.get("variables", {})

        # Some variables text can be multiline, so we need to replace newlines with spaces
        variables = {key: value.replace("\n", " ") for key, value in variables.items()}

        # Create jinja2 template from original yaml and render it using variables
        template = Template(yaml_content)
        rendered_yaml = template.render(variables)

        # Finally, load rendered yaml
        yaml_dict = yaml.safe_load(rendered_yaml)
    else:
        yaml_dict = yaml.safe_load(yaml_content)

    return yaml_dict



from enum import StrEnum

from pydantic import BaseModel

from contextcheck.assertions.llm_eval_prompts import (
    HALLUCINATION_PROMPT_TEMPLATE,
    HUMAN_VS_AI_PROMPT_TEMPLATE,
    MODEL_GRADING_QA_PROMPT_TEMPLATE,
    QA_REFERENCE_PROMPT_TEMPLATE,
    SUMMARIZATION_PROMPT_TEMPLATE,
)
from contextcheck.endpoints.endpoint import EndpointBase
from contextcheck.models.request import RequestBase
from contextcheck.models.response import ResponseBase


class LlmMetricEnum(StrEnum):
    HALLUCINATION = "hallucination"
    QA_REFERENCE = "qa-reference"
    MODEL_GRADING_QA = "model-grading-qa"
    SUMMARIZATION = "summarization"
    HUMAN_VS_AI = "human-vs-ai"


class LLMMetric(BaseModel):
    prompt_template: str
    rails: dict = {}

    def parse_prompt(self, input: str, output: str, **kwargs) -> str:
        return self.prompt_template.format(
            input=input, output=output, reference=kwargs["reference"]
        )

    def check_response_rails(self, response: ResponseBase) -> bool:
        if response.message.lower() not in self.rails:  # type: ignore
            raise ValueError(
                f"Response message '{response.message}' should be one of : {list(self.rails.keys())}"
            )

        return self.rails[response.message.lower()]  # type: ignore


class MetricHallucination(LLMMetric):
    prompt_template: str = HALLUCINATION_PROMPT_TEMPLATE
    rails: dict = {"factual": True, "hallucinated": False}


class MetricQAReference(LLMMetric):
    prompt_template: str = QA_REFERENCE_PROMPT_TEMPLATE
    rails: dict = {"correct": True, "incorrect": False}


class MetricModelGradingQA(LLMMetric):
    prompt_template: str = MODEL_GRADING_QA_PROMPT_TEMPLATE
    rails: dict = {"correct": True, "incorrect": False}

    def parse_prompt(self, input: str, output: str, **kwargs) -> str:
        return self.prompt_template.format(output=output, assertion=kwargs["assertion"])


class MetricSummarization(LLMMetric):
    prompt_template: str = SUMMARIZATION_PROMPT_TEMPLATE
    rails: dict = {"good": True, "bad": False}

    def parse_prompt(self, input: str, output: str, **kwargs) -> str:
        return self.prompt_template.format(input=input, output=output)


class MetricHumanVsAI(LLMMetric):
    prompt_template: str = HUMAN_VS_AI_PROMPT_TEMPLATE
    rails: dict = {"correct": True, "incorrect": False}


LLM_METRIC_TYPE_MAPPING = {
    LlmMetricEnum.HALLUCINATION: MetricHallucination,
    LlmMetricEnum.QA_REFERENCE: MetricQAReference,
    LlmMetricEnum.MODEL_GRADING_QA: MetricModelGradingQA,
    LlmMetricEnum.SUMMARIZATION: MetricSummarization,
    LlmMetricEnum.HUMAN_VS_AI: MetricHumanVsAI,
}


def llm_metric_factory(metric_type: LlmMetricEnum) -> LLMMetric:
    try:
        metric_class = LLM_METRIC_TYPE_MAPPING[metric_type]
    except KeyError:
        raise ValueError(f"No metric found for type '{metric_type}'")

    return metric_class()


class LLMMetricEvaluator(BaseModel):
    eval_endpoint: EndpointBase
    metric: LLMMetric

    class EvalRequest(RequestBase):
        message: str = ""

    def evaluate(self, input: str, output: str, **kwargs) -> bool:
        message = self.metric.parse_prompt(input=input, output=output, **kwargs)
        response = self.eval_endpoint.send_request(self.EvalRequest(message=message))
        return self.metric.check_response_rails(response)



from loguru import logger

from contextcheck.endpoints.factory import factory as endpoint_factory
from contextcheck.interfaces.interface import InterfaceBase
from contextcheck.models.models import TestScenario, TestStep


class Executor:
    """
    Executes a given scenario.

    The idea here is to use executor to have all the logic needed for conducting test scenario
    basically what to "do" with the TestScenario object.
    Thus, Test* models represent mainly data handling, parsing etc.
    Doing so allows for flexible test logic definition with debugging, different outputs, concurrency etc.
    Even distributed tests later on.
    """

    def __init__(
        self,
        test_scenario: TestScenario,
        ui: InterfaceBase | None = None,
        exit_on_failure: bool = False,
    ) -> None:
        self.test_scenario = test_scenario
        self.context: dict = {}
        self.ui = ui or InterfaceBase()
        self.exit_on_failure = exit_on_failure
        self.endpoint_under_test = endpoint_factory(self.test_scenario.config.endpoint_under_test)
        self.eval_endpoint = (
            endpoint_factory(self.test_scenario.config.eval_endpoint)
            if self.test_scenario.config.eval_endpoint
            else None
        )
        self.early_stop = False

    def run_all(self) -> bool | None:
        """Run all test steps sequentially."""
        logger.info("Running scenario", self.test_scenario)
        result = True
        for test_step in self.run_steps():
            result &= bool(test_step.result)
            if self.early_stop:
                break
        self.test_scenario.result = result
        return result

    def run_steps(self) -> list[TestStep]:
        """Run all steps and return them."""
        step_results = []
        for test_step in self.test_scenario.steps:
            if self.early_stop:
                break
            step = self._run_step(test_step)
            step_results.append(step)
        return step_results

    def _run_step(self, test_step: TestStep) -> TestStep:
        """Run a given step and update result."""

        self.ui(test_step)

        request = test_step.request.build(self.context)

        self.ui(request)

        response = self.endpoint_under_test.send_request(request)
        test_step.response = response
        self._update_context(last_response=response)

        self.ui(response)

        result = True
        for assertion in test_step.asserts:
            try:
                result &= assertion(request, response, eval_endpoint=self.eval_endpoint)
            except Exception as e:
                logger.error(f"Error during assertion: {e}")
                result = False
            self.ui(assertion)
            if result is False and self.exit_on_failure:
                self.early_stop = True
                break
        test_step.result = result
        return test_step

    def _update_context(self, **data) -> None:
        """Update executor context to store global execution data."""
        self.context.update(data)

    def summary(self, **kwargs: dict[str, str]):
        self.ui.summary(self, **kwargs)

    def report_results(self, **kwargs):
        # This function isn't utilized on it's own right now, but is rather used by summary
        # at least for tui interface, though I leave it as a separate accessible function
        self.ui.report_results(self, **kwargs)

    def report_time(self, **kwargs):
        self.ui.report_time(self, **kwargs)



from __future__ import annotations

import copy
from collections import defaultdict
from typing import TYPE_CHECKING, Any

import numpy as np
from loguru import logger
from pydantic import BaseModel

from contextcheck.assertions.settings import AssertionKind

if TYPE_CHECKING:
    from contextcheck.executors.executor import Executor


class InterfaceBase(BaseModel):
    """UI should "know" what to do with a given object."""

    def __call__(self, obj: Any) -> Any:
        logger.info(obj)

    @staticmethod
    def summary(obj: Any, **kwargs: Any) -> Any:
        logger.info(obj)

    def _create_a_summary_report(
        self, executor: Executor
    ) -> dict[str, dict[str, dict[str, float]]]:
        """
        Create a summary of the results obtained from a test scenario

        Warning: The results must be a nested dict of type dict[str, dict[str, dict[str, float]]]
        """

        # This function is taken straight from assertion factory
        def get_assertion_kind(assertion: dict):
            kind = next(
                assert_key
                for assert_key in assertion.keys()
                if assert_key in [ak.value for ak in AssertionKind]
            )
            return kind

        # Summarize binary statistics
        def summarize_binary_statistics(data: list[float]) -> dict[str, float]:
            if not data:
                return {"mean": 0.0, "count": 0}

            count1 = sum(data)
            count = len(data)
            mean = count1 / count

            return {"mean": mean, "count": count}

        # In place aggregation of the assertion results
        def aggregate_data(data: dict):
            for key, value in data.items():
                if isinstance(value, dict):
                    aggregate_data(value)
                else:
                    # NOTE: Currently all statistics are binary, therefore it can be leave as it is
                    # If continous statistics occur then this functionality shall be extended
                    # NOTE: One option that I see is to create (key-type) mapper where each
                    # statistic would be summarized differently depending on its name
                    data[key] = summarize_binary_statistics(data[key])

        assertion_results = defaultdict(lambda: defaultdict(list))
        for step in executor.test_scenario.steps:
            for assertion in step.asserts:
                assertion = assertion.model_dump()
                kind = get_assertion_kind(assertion=assertion)
                assertion_results[kind][
                    assertion[kind] if kind != AssertionKind.EVAL else kind
                ].append(
                    float(assertion["result"]) if assertion["result"] is not None else 0.0
                )  # Dump results to float for easier manipulation

        # Copy is technically not necessary, but it logically separates the results from
        # the aggregation
        assertion_results_aggregated = copy.deepcopy(assertion_results)
        aggregate_data(data=assertion_results_aggregated)
        return assertion_results_aggregated

    def _create_time_statistics(self, executor: "Executor") -> dict[str, float]:
        request_times = []
        for step in executor.test_scenario.steps:
            request_times.append(step.response.stats.conn_duration)

        mean = np.mean(request_times)
        median = np.median(request_times)
        minimum = np.min(request_times)
        maximum = np.max(request_times)
        std = np.std(request_times)

        time_statistics = {
            "mean": mean,
            "median": median,
            "minimum": minimum,
            "maximum": maximum,
            "std": std,
        }
        return time_statistics

    def report_results(self, executor: "Executor", **kwargs) -> None:
        raise NotImplementedError("Functionality is not implemented")

    def report_time(self, executor: "Executor", **kwargs) -> None:
        raise NotImplementedError("Functionality is not implemented")
